"""River Avulsion Module."""

from ._version import get_versions
from .riverbmi import BmiRiverModule
from .rivermodule import RiverModule

__all__ = ["BmiRiverModule", "RiverModule"]

__version__ = get_versions()["version"]
del get_versions



#! /usr/local/bin/python
# -*- coding: utf-8 -*-

import numpy as np


def add_to_neighboring_cells(z, sub, inc, win=1):
    """Add a value to all neighboring cells.

    Parameters
    ----------
    z : ndarray
        2D array of values.
    sub : tuple of int
        Row/column subscripts into array.
    inc : float
        Value to increment *z* by.
    win : int
        Size of the window around *sub*.

    Examples
    --------
    >>> x = np.zeros((4, 5))
    >>> add_around_cell(x, (0, 0), 1)
    """
    z[
        max(0, sub[0] - win) : min(z.shape[0], sub[0] + win + 1),
        max(0, sub[1] - win) : min(z.shape[1], sub[1] + win + 1),
    ] += inc


def dep_blanket(current_SL, blanket_rate, n, riv_i, riv_j, ch_depth):
    depo_flag = np.ones(n.shape, dtype=np.int)

    # don't deposit in the river course
    depo_flag[riv_i, riv_j] = 0

    # don't deposit if cell <= sea level
    # ASK BRAD ABOUT THIS (no cell is below SL right now)
    # but it might be when couple?? need to figure out if keeping up
    # with elevation RELATIVE to sea level or actual elevation??
    depo_flag[n <= current_SL] = 0

    # if cell elevation is above bankfull elev, don't deposit
    bankfull_elevation = n[riv_i, riv_j] + ch_depth
    for row in riv_i:
        depo_flag[row, n[row] >= bankfull_elevation[row]] = 0

    # don't deposit on first two rows b/c inlet rise rate does that
    depo_flag[:2, :] = 0

    # deposit "blanket" deposition on qualified cells
    n[depo_flag == 1] += blanket_rate

    # dn_fp = depo_flag * blanket_rate

    # return n, dn_fp


def distance_to_river(y, y0):
    return np.absolute(y - y0)


def within_wetland(y, riv_ind, wetland_width=0.0):
    dy = distance_to_river(y, y[riv_ind])
    is_wetland = dy <= wetland_width
    is_wetland[riv_ind] = False
    return is_wetland


def wetlands(current_SL, WL_Z, wetland_width, n, riv_i, riv_j, y, x):
    depo_wetland = np.zeros(n.shape, dtype=np.int)

    for row, col in zip(riv_i, riv_j):
        dist = within_wetland(y[row], col, wetland_width=wetland_width)
        elev = n[row] < current_SL + WL_Z

        cols = dist & elev & (depo_wetland[row] == 0)

        n[row, cols] = current_SL + WL_Z

        depo_wetland[row, cols] == 1


def dep_splay(n, ij_fail, splay_dep, splay_type=1):
    """Deposit around a failed river cell.

    Parameters
    ----------
    n : ndarray
        Elevation array.
    ij_path : tuple of int
        Row and column of the river failure.
    old_path : tuple of array_like
        Row and column indices for the old river path.
    a : int
        River path index of the failure.
    splay_depth : float
        Deposition depth.
    splay_type : {1, 2}, optional
        Failure type

    USE DEPTH-DEPENDENCE IN THE FUTURE
        SE1 = (n[riv_x[a]/dx][riv_y[a]/dy] + ch_depth - \
                n[new_riv_x[a]/dx][new_riv_y[a]/dy]) / ch_depth

    ADD LARGER SPLAY TYPE IN FUTURE?
    (could be first two failed river cells + surrounding)

    This could possibly be improved by comparing to find nearest beach routine (CEM)
    or using some sort of search radius
    """

    if splay_type == 1:  # splay deposition just at first failed river cell
        n[ij_fail] += splay_dep
    if splay_type == 2:  # splay deposition at first failed river cell
        # and the adjacent cells
        add_to_neighboring_cells(n, ij_fail, splay_dep)


def dep_fines(n, riv_i, riv_j, dn_rc, frac_fines, SL):

    fine_dep = np.zeros_like(n)
    dn_rc = np.insert(dn_rc, [0], 0)
    dn_rc = np.append(dn_rc, dn_rc[-1])

    for k in range(1, len(riv_i)):

        dep_rate = frac_fines * dn_rc[k]

        if dep_rate > 0:
            if riv_j[k] == 0 and riv_i[k] == 0:
                di, dj = np.array([1, 1, 0]), np.array([0, 1, 1])
            elif riv_j[k] == 0 and riv_i[k] == n.shape[0] - 1:
                di, dj = np.array([-1, -1, 0]), np.array([0, 1, 1])
            elif riv_j[k] == n.shape[1] - 1 and riv_i[k] == 0:
                di, dj = np.array([0, 1, 1]), np.array([-1, -1, 0])
            elif riv_j[k] == n.shape[1] - 1 and riv_i[k] == n.shape[0] - 1:
                di, dj = np.array([0, -1, -1]), np.array([-1, -1, 0])
            elif riv_j[k] == n.shape[1] - 1:
                di, dj = np.array([-1, -1, 0, 1, 1]), np.array([0, -1, -1, -1, 0])
            elif riv_j[k] == 0:
                di, dj = np.array([-1, -1, 0, 1, 1]), np.array([0, 1, 1, 1, 0])
            elif riv_i[k] == n.shape[0] - 1:
                di, dj = np.array([0, -1, -1, -1, 0]), np.array([-1, -1, 0, 1, 1])
            elif riv_i[k] == 0:
                di, dj = np.array([0, 1, 1, 1, 0]), np.array([-1, -1, 0, 1, 1])
            else:
                di, dj = (
                    np.array([0, -1, -1, -1, 0, 1, 1, 1]),
                    np.array([-1, -1, 0, 1, 1, 1, 0, -1]),
                )

            for m in range(len(di)):
                if (fine_dep[riv_i[k] + di[m], riv_j[k] + dj[m]] < dep_rate) and (
                    n[riv_i[k] + di[m], riv_j[k] + dj[m]] > SL
                ):
                    fine_dep[riv_i[k] + di[m], riv_j[k] + dj[m]] = dep_rate

    fine_dep[riv_i, riv_j] = 0
    n += fine_dep

    return



#! /usr/local/bin/python
import numpy as np
import yaml
from scipy.ndimage import measurements


def read_params_from_file(fname):
    """Read model parameters from a file.

    Parameters
    ----------
    fname : str
        Name of YAML-formatted parameters file.

    Returns
    -------
    dict
        A dict of parameters for the heat model.
    """
    with open(fname, "r") as fp:
        params = yaml.safe_load(fp)

    return params


def is_diagonal_neighbor(sub0, sub1):
    return (sub0[0] != sub1[0]) and (sub0[1] != sub1[1])


def is_same_row(sub0, sub1):
    return sub0[0] == sub1[0]


def channel_is_superelevated(z, riv, behind, channel_depth, super_ratio, sea_level):
    """Check if a channel location is super-elevated.

    Parameters
    ----------
    z : ndarray
        2D-array of elevations.
    sub : tuple of int
        Row-column subscripts into *z*.
    channel_depth : float
        The depth of the channel.
    super_ratio : float
        Ratio of bankfull elevation to out-of-channel elevation for
        channel to be super-elevated.

    Returns
    -------
    boolean
        `True` if channel is super-elevated. Otherwise, `False`.
    """
    superelev = 0
    z_bankfull = z[riv] + channel_depth

    # cross-shore river orientation
    if behind[0] + 1 == riv[0]:
        if riv[1] == 0:
            adj1 = z[riv[0], riv[1] + 1]
            adj2 = z[riv[0], riv[1] + 1]  # not sure if this needs to be included
        elif riv[1] == z.shape[1] - 1:
            adj1 = z[riv[0], riv[1] - 1]
            adj2 = z[riv[0], riv[1] - 1]
        else:
            adj1, adj2 = z[riv[0], riv[1] - 1], z[riv[0], riv[1] + 1]

    # alongshore river orientation
    if behind[0] == riv[0]:
        if riv[0] == z.shape[0] - 1:
            adj1 = z[riv[0] + 1, riv[1]]
            adj2 = z[riv[0] + 1, riv[1]]
        else:
            adj1, adj2 = z[riv[0] + 1, riv[1]], z[riv[0] - 1, riv[1]]

    threshold = super_ratio * channel_depth

    if (adj1 < sea_level) and (adj2 < sea_level):
        superelev = 0
    elif adj1 < sea_level:
        if z_bankfull - adj2 >= threshold:
            superelev = 1
    elif adj2 < sea_level:
        if z_bankfull - adj1 >= threshold:
            superelev = 1
    else:
        if ((z_bankfull - adj2) or (z_bankfull - adj1)) >= threshold:
            superelev = 1

    return superelev


def get_link_lengths(path, dx=1.0, dy=1.0):
    DIAGONAL_LENGTH = np.sqrt(dx ** 2.0 + dy ** 2.0)

    lengths = np.empty(len(path[0]) - 1, dtype=np.float)
    ij_last = path[0][0], path[1][0]
    for n, ij in enumerate(zip(path[0][1:], path[1][1:])):
        if is_diagonal_neighbor(ij, ij_last):
            lengths[n] = DIAGONAL_LENGTH
        elif is_same_row(ij, ij_last):
            lengths[n] = dx
        else:
            lengths[n] = dy
        ij_last = ij
    return lengths


def get_channel_distance(path, dx=1.0, dy=1.0):
    total_distance = get_link_lengths(path, dx=dx, dy=dy).cumsum()
    return np.append(0, total_distance)


def find_path_length(n, path, sea_level, ch_depth, slope, dx=1.0, dy=1.0):
    beach_len = find_new_beach_length(
        n,
        (path[0][-2], path[1][-2]),
        (path[0][-1], path[1][-1]),
        sea_level,
        dx=dx,
        dy=dy,
    )

    lengths = get_link_lengths(path, dx=dx, dy=dy)
    lengths[-1] = np.divide(lengths[-1], 2.0) + beach_len
    riv_length = lengths.sum()

    return riv_length


def find_riv_path_length(n, path, sea_level, ch_depth, slope, dx=1.0, dy=1.0):
    beach_len = find_beach_length_riv_cell(
        n,
        (path[0][-2], path[1][-2]),
        (path[0][-1], path[1][-1]),
        sea_level,
        ch_depth,
        slope,
        dx=dx,
        dy=dy,
    )

    lengths = get_link_lengths(path, dx=dx, dy=dy)
    lengths[-1] = np.divide(lengths[-1], 2.0) + beach_len
    riv_length = lengths.sum()

    return riv_length


def find_point_in_path(path, sub):
    """Find a point in a path.

    Parameters
    ----------
    path : tuple of array_like
        Tuple of arrays of int that represent indices into a matrix.
    sub : tuple in int
        Indices to search *path* for.

    Returns
    -------
    int or None
        Index into *path* if the indices are found. Otherwise, ``None``.

    Examples
    --------
    >>> path = ((0, 1, 4, 6, 5), (3, 1, 7, 8, 10))
    >>> find_point_in_path(path, (4, 7))
    2
    >>> find_point_in_path(path, (99, 2)) is None
    True
    """
    try:
        return list(zip(*path)).index(sub)
    except ValueError:
        return None


def set_linear_profile(z, riv_ij, dx=1.0, dy=1.0):
    """Set elevations along a path to be linear.

    Examples
    --------
    >>> import numpy as np
    >>> z = np.array([[0, 0, 0, 0, 0],
    ...               [1, 2, 1, 1, 1],
    ...               [2, 2, 2, 2, 2],
    ...               [3, 3, 3, 3, 3]], dtype=float)
    >>> ij = [(0, 2), (1, 2), (2, 2), (3, 2)]
    >>> set_linear_profile(z, ij)
    array([[ 0.,  0.,  0.,  0.,  0.],
           [ 1.,  2.,  1.,  1.,  1.],
           [ 2.,  2.,  2.,  2.,  2.],
           [ 3.,  3.,  3.,  3.,  3.]])

    >>> ij = [(0, 0), (1, 1), (2, 2), (3, 1)]
    >>> set_linear_profile(z, ij, dx=3., dy=4.)
    array([[ 0.,  0.,  0.,  0.,  0.],
           [ 1.,  1.,  1.,  1.,  1.],
           [ 2.,  2.,  2.,  2.,  2.],
           [ 3.,  3.,  3.,  3.,  3.]])

    """
    z[zip(*riv_ij[:-1])] = z[riv_ij[-1]] + np.arange(len(riv_ij) - 1, 0, -1) * 1e-6

    return z


def set_linear_slope(z, riv_ij, dx=1.0, dy=1.0):
    """Set slope along a path to be linear."""

    if len(riv_ij) > 1:
        z0 = z[riv_ij[0]]
        dz = z[riv_ij[-1]] - z[riv_ij[0]]

        lengths = get_link_lengths(zip(*riv_ij), dx=dx, dy=dy)
        ds = lengths.sum()

        z[zip(*riv_ij[1:])] = z0 + dz / ds * lengths.cumsum()

    return z


def fill_upstream(z, riv_ij, dx=1.0, dy=1.0):
    """Fill depressions upstream of a pit.

    Exammples
    ---------
    >>> import numpy as np
    >>> z = np.array([[3, 3, 3, 3, 3],
    ...               [2, 2, 2, 2, 2],
    ...               [1, 1, 0, 1, 1],
    ...               [0, 0, 1, 0, 0]], dtype=float)
    >>> ij = [(0, 2), (1, 2), (2, 2), (3, 2)]
    >>> fill_upstream(z, ij)
    array([[ 3. ,  3. ,  3. ,  3. ,  3. ],
           [ 2. ,  2. ,  2. ,  2. ,  2. ],
           [ 1. ,  1. ,  1.5,  1. ,  1. ],
           [ 0. ,  0. ,  1. ,  0. ,  0. ]])

    >>> z = np.array([[3, 3, 1, 3, 3],
    ...               [2, 2, 0, 2, 2],
    ...               [1, 1, 1, 1, 1],
    ...               [0, 0, 2, 0, 0]], dtype=float)
    >>> ij = [(0, 2), (1, 2), (2, 2), (3, 2)]
    >>> fill_upstream(z, ij)
    """
    fill_to = z[riv_ij[-1]]

    for n in range(len(riv_ij) - 2, -1, -1):
        if z[riv_ij[n]] > fill_to:
            break
        # z[riv_ij[n]] = z[riv_ij[n + 1]] + 1e-6

    if z[riv_ij[n]] <= fill_to:
        z[riv_ij[n]] = fill_to + 1e-6

    set_linear_profile(z, riv_ij[n:], dx=dx, dy=dy)

    return z


def sort_lowest_neighbors(n, sub):
    """Sort neighbor cells by elevation."""
    i, j = sub

    if j == n.shape[1] - 1:
        di, dj = np.array([0, 1, 1]), np.array([-1, -1, 0])
    elif j == 0:
        di, dj = np.array([1, 1, 0]), np.array([0, 1, 1])
    else:
        di, dj = np.array([0, 1, 1, 1, 0]), np.array([-1, -1, 0, 1, 1])

    sorted = np.argsort(n[i + di, j + dj])

    return i + di[sorted], j + dj[sorted]


def find_new_beach_length(n, sub0, sub1, sea_level, dx=1.0, dy=1.0):

    cell_elev = n[sub1] - sea_level

    DIAGONAL_LENGTH = np.sqrt(dx ** 2.0 + dy ** 2.0)

    if is_diagonal_neighbor(sub0, sub1):
        d_dist = DIAGONAL_LENGTH

    elif is_same_row(sub0, sub1):
        d_dist = dx

    else:
        d_dist = dy

    return cell_elev * d_dist


def find_beach_length_riv_cell(
    n, sub0, sub1, sea_level, channel_depth, slope, dx=1.0, dy=1.0
):
    """Find length of beach in shoreline cell with river.

    Parameters
    ----------
    z : ndarray
        2D-array of elevations.
    sub0 : tuple of int
        Row-column subscripts into *z*. (cell prior to beach)
    sub1 : tuple of int
        Row-column subscripts into *z*. (beach cell)
    sea_level : float
        Elevation of current sea level.
    channel_depth : float
        The depth of the channel.

    Returns
    -------
    Length of beach in a cell. """

    cell_elev = n[sub1] + channel_depth - sea_level

    max_cell_h = slope * dx

    DIAGONAL_LENGTH = np.sqrt(dx ** 2.0 + dy ** 2.0)

    if is_diagonal_neighbor(sub0, sub1):
        d_dist = DIAGONAL_LENGTH

    elif is_same_row(sub0, sub1):
        d_dist = dx

    else:
        d_dist = dy

    return (cell_elev / max_cell_h) * d_dist


def lowest_cell_elev(n, sub):
    i, j = sub

    if j == 0 and i == 0:
        di, dj = np.array([1, 1, 0]), np.array([0, 1, 1])
    elif j == 0 and i == n.shape[0] - 1:
        di, dj = np.array([-1, -1, 0]), np.array([0, 1, 1])
    elif j == n.shape[1] - 1 and i == 0:
        di, dj = np.array([0, 1, 1]), np.array([-1, -1, 0])
    elif j == n.shape[1] - 1 and i == n.shape[0] - 1:
        di, dj = np.array([0, -1, -1]), np.array([-1, -1, 0])
    elif j == n.shape[1] - 1:
        di, dj = np.array([-1, -1, 0, 1, 1]), np.array([0, -1, -1, -1, 0])
    elif j == 0:
        di, dj = np.array([-1, -1, 0, 1, 1]), np.array([0, 1, 1, 1, 0])
    elif i == n.shape[0] - 1:
        di, dj = np.array([0, -1, -1, -1, 0]), np.array([-1, -1, 0, 1, 1])
    elif i == 0:
        di, dj = np.array([0, 1, 1, 1, 0]), np.array([-1, -1, 0, 1, 1])
    else:
        di, dj = (
            np.array([0, -1, -1, -1, 0, 1, 1, 1]),
            np.array([-1, -1, 0, 1, 1, 1, 0, -1]),
        )

    lowest = np.amin(n[i + di, j + dj])

    return lowest


def lowest_face(n, sub):
    i, j = sub

    if j == 0 and i == 0:
        di, dj = np.array([1, 0]), np.array([0, 1])
    elif j == 0 and i == n.shape[0] - 1:
        di, dj = np.array([-1, 0]), np.array([0, 1])
    elif j == n.shape[1] - 1 and i == 0:
        di, dj = np.array([0, 1]), np.array([-1, 0])
    elif j == n.shape[1] - 1 and i == n.shape[0] - 1:
        di, dj = np.array([0, -1]), np.array([-1, 0])
    elif j == n.shape[1] - 1:
        di, dj = np.array([-1, 0, 1]), np.array([0, -1, 0])
    elif j == 0:
        di, dj = np.array([-1, 0, 1]), np.array([0, 1, 0])
    elif i == n.shape[0] - 1:
        di, dj = np.array([0, -1, 0]), np.array([-1, 0, 1])
    elif i == 0:
        di, dj = np.array([0, 1, 0]), np.array([-1, 0, 1])
    else:
        di, dj = np.array([0, -1, 0, 1]), np.array([-1, 0, 1, 0])

    lowest_face = np.amin(n[i + di, j + dj])

    return lowest_face


def fix_elevations(z, riv_i, riv_j, ch_depth, sea_level, slope, dx, max_rand, SLRR):

    test_elev = z - sea_level
    max_cell_h = slope * dx
    riv_prof = test_elev[riv_i, riv_j]
    test_elev[riv_i, riv_j] += 2 * ch_depth

    # set new subaerial cells to marsh elevation
    test_elev[test_elev == 0] = max_cell_h

    # make mask for depressions
    ocean_mask = test_elev < max_cell_h
    labeled_ponds, ocean = measurements.label(ocean_mask)

    # # fill in underwater spots that are below SL (?)
    # below_SL = [z <= sea_level]
    # underwater_cells, big_ocean = measurements.label(below_SL)
    # underwater_cells[underwater_cells == big_ocean] = 0
    # test_elev[underwater_cells > 0] = max_cell_h + SLRR + (np.random.rand() * max_rand)

    # create an ocean and shoreline mask
    ocean_and_shore = np.copy(labeled_ponds)

    # create an ocean mask
    # ocean_cells = np.copy(ocean_and_shore)
    # ocean_and_shore[test_elev > 0] = 0

    # create mask for pond cells and fix them
    area = measurements.sum(
        ocean_mask, labeled_ponds, index=np.arange(labeled_ponds.max() + 1)
    )
    areaPonds = area[labeled_ponds]
    labeled_ponds[areaPonds == areaPonds.max()] = 0

    # finish creating ocean and shoreline mask
    ocean_and_shore[areaPonds != areaPonds.max()] = 0

    # something here to get rid of ocean cells
    test_elev[labeled_ponds > 0] = max_cell_h + SLRR + (np.random.rand() * max_rand)

    # raise cells close to sea level above it
    test_elev[(test_elev >= max_cell_h) & (test_elev <= (max_cell_h + SLRR))] = (
        max_cell_h + SLRR + (np.random.rand() * max_rand)
    )

    riv_buffer = np.zeros_like(test_elev)
    riv_buffer[riv_i, riv_j] = 1
    riv_buffer[riv_i[1:] - 1, riv_j[1:]] = 1

    for i in range(1, test_elev.shape[0] - 1):
        for j in range(test_elev.shape[1]):
            if (
                not ocean_and_shore[i, j]
                and not ocean_and_shore[i - 1, j]
                and not ocean_and_shore[i + 1, j]
                and not riv_buffer[i, j]
            ):
                if test_elev[i + 1, j] >= test_elev[i, j]:
                    test_elev[i, j] = test_elev[i + 1, j] + (np.random.rand() * slope)

    test_elev[riv_i, riv_j] = riv_prof

    z = test_elev + sea_level

    return z


def fill_abandoned_channel(
    breach_loc, n, new, riv_i, riv_j, current_SL, ch_depth, slope, dx
):

    new_profile = n[new[0], new[1]]

    max_shorecell_h = current_SL + (slope * dx)
    riv_cells = np.zeros_like(n)
    riv_cells[riv_i, riv_j] = 1
    riv_cells[new[0], new[1]] = 1

    n[riv_i[-1], riv_j[-1]] += ch_depth

    if len(riv_i) - breach_loc > 1:
        for k in range(breach_loc, len(riv_i) - 1):
            if (riv_j[k] == n.shape[1] - 1) or (riv_j[k] == 0):
                n[riv_i[k], riv_j[k]] = max_shorecell_h + (np.random.rand() * slope)
            elif (
                riv_cells[riv_i[k], riv_j[k] + 1] and riv_cells[riv_i[k], riv_j[k] - 1]
            ):
                n[riv_i[k], riv_j[k]] = max_shorecell_h + (np.random.rand() * slope)
            elif riv_cells[riv_i[k], riv_j[k] + 1]:
                if n[riv_i[k], riv_j[k] - 1] <= max_shorecell_h:
                    n[riv_i[k], riv_j[k]] = max_shorecell_h + (np.random.rand() * slope)
                else:
                    n[riv_i[k], riv_j[k]] = n[riv_i[k], riv_j[k] - 1] + (
                        np.random.rand() * slope
                    )
            elif riv_cells[riv_i[k], riv_j[k] - 1]:
                if n[riv_i[k], riv_j[k] + 1] <= max_shorecell_h:
                    n[riv_i[k], riv_j[k]] = max_shorecell_h + (np.random.rand() * slope)
                else:
                    n[riv_i[k], riv_j[k]] = n[riv_i[k], riv_j[k] + 1] + (
                        np.random.rand() * slope
                    )
            elif (n[riv_i[k], riv_j[k] + 1] <= max_shorecell_h) or (
                n[riv_i[k], riv_j[k] - 1] <= max_shorecell_h
            ):
                n[riv_i[k], riv_j[k]] = max_shorecell_h + (np.random.rand() * slope)
            else:
                n[riv_i[k], riv_j[k]] = (
                    (n[riv_i[k], riv_j[k] + 1] + n[riv_i[k], riv_j[k] - 1]) / 2
                ) + (np.random.rand() * slope)

    n[new[0], new[1]] = new_profile



#! /usr/local/bin/python
# -*- coding: utf-8 -*-
from .avulsion_utils import find_new_beach_length, get_link_lengths


def cut_init(riv_i, riv_j, n, init_cut):
    n[riv_i, riv_j] -= init_cut

    return n


def cut_new(riv_i, riv_j, n, sea_level, ch_depth, dx=1.0, dy=1.0):
    """Set elevations of new portions of a profile."""

    if riv_i.size > 1:
        beach_len = find_new_beach_length(
            n, (riv_i[-2], riv_j[-2]), (riv_i[-1], riv_j[-1]), sea_level, dx=dx, dy=dy
        )

        lengths = get_link_lengths((riv_i, riv_j), dx=dx, dy=dy)
        lengths[-1] += beach_len

        i0, j0 = riv_i[0], riv_j[0]
        z0 = n[riv_i[0], riv_j[0]]

        # calculate slope of new stretch of river
        # slope = (n[i0, j0] - n[riv_i[-1], riv_j[-1]]) / lengths.sum()
        slope = (n[i0, j0] - (n[riv_i[-1], riv_j[-1]] - ch_depth)) / lengths.sum()

        n[riv_i[1:], riv_j[1:]] = z0 - slope * lengths.cumsum()


def cut_local(riv_i, riv_j, n, dx=1.0, dy=1.0):
    """Set elevations of new portions of a profile after local avulsion."""

    if riv_i.size > 1:

        lengths = get_link_lengths((riv_i, riv_j), dx=dx, dy=dy)

        i0, j0 = riv_i[0], riv_j[0]
        z0 = n[riv_i[0], riv_j[0]]

        # calculate slope of new stretch of river
        slope = (n[i0, j0] - n[riv_i[-1], riv_j[-1]]) / lengths.sum()

        n[riv_i[1:], riv_j[1:]] = z0 - slope * lengths.cumsum()



#! /usr/local/bin/python
# -*- coding: utf-8 -*-


def elev_change(current_SL, n, riv_i, riv_j, ch_depth, SLRR):
    """Raise elevations to sea level if below sea level.

    Set elevations of cells below sea level to sea level unless they are
    within the river channel.

    Parameters
    ----------
    current_SL : float
        Sea level.
    n : ndarray
        Array of elevations.
    riv_i : ndarray of int
        Row indices into *n* for river cells.
    riv_j : ndarray of int
        Column indices into *n* for river cells.
    ch_depth : float
        Channel depth.
    """
    # changes elevation of last river course cell according to sea level change
    # n[riv_i[-1], riv_j[-1]] = current_SL - ch_depth

    n[riv_i[-1], riv_j[-1]] += SLRR

    channel_elevations = n[riv_i, riv_j].copy()

    # raises cell elevation to sea level if it is below
    # n[n < current_SL] = current_SL
    # channel_elevations[channel_elevations < current_SL] = current_SL

    n[riv_i, riv_j] = channel_elevations

    # Need to somehow change above treatment of cells... they don't need to be
    # raised to sea level anymore.


#    # raises elevation of whole inlet row
#    for I in range(jmax):
#        n[0][I] = n[0][I] + (IRR)
#        I = I + 1



#! /usr/local/bin/python

import numpy as np


def linear_subsidence(n, riv_i, riv_j, ch_depth, sub_rate, sub_start, SL):
    """ Subside cells in rows beyond start location by linear rate.
    Updated version could be to subside at an increasing rate towards ocean. """

    subside_cells = np.zeros_like(n)
    subside_cells[sub_start:, :] = 1

    subaerial_elev = np.copy(n)
    subaerial_elev[riv_i, riv_j] += ch_depth

    subside_cells[subaerial_elev <= SL] = 0

    n[subside_cells == 1] -= sub_rate



#! /usr/local/bin/python

import errno
import os

import numpy as np
import yaml

from . import FP, avulse, avulsion_utils, diffuse, downcut, flux, steep_desc, subside
from .avulsion_utils import read_params_from_file

_SECONDS_PER_YEAR = 31536000.0
_SECONDS_PER_DAY = 86400.0


def make_empty_file(path):
    """Create an empty file.

    Create an empty file along with all of its parent folders,
    if necessary. Note that if the file already exists, it
    will be clobbered.

    Parameters
    ----------
    path : str
        Path to the file to create.
    """
    dirname = os.path.dirname(path)
    try:
        os.makedirs(dirname)
    except OSError as err:
        if err.errno == errno.EEXIST and os.path.isdir(dirname):
            pass
        else:
            raise

    with open(path, "w"):
        pass


class RiverModule(object):
    def __init__(
        self,
        rand_seed=1945,
        spacing=(0.1, 0.1),
        shape=(100, 120),
        n0=5.0,
        nslope=0.001,
        max_rand=0.1,
        dt_day=0.05,
        Initial_SL=0.0,
        SLRR_m=0.0,
        SubRate_m=0.0,
        Sub_Start=0,
        ch_width=10.0,
        ch_depth=1.0,
        ch_discharge=10.0,
        A=1.0,
        c_f=0.01,
        C_0=1.0,
        sed_sg=2.65,
        init_cut_frac=1.0,
        super_ratio=1.0,
        short_path=1,
        WL_Z=0.0,
        WL_dist=0,
        blanket_rate_m=0.0,
        fine_dep_frac=0.0,
        splay_type=2,
        saveavulsions=False,
        savecourseupdates=False,
    ):
        """The RAFEM.

        Parameters
        ----------
        rand_seed : int
            Seed for random number generator.
        spacing : tuple of float
            Row (dy) and column (dx) spacing [m].
        shape : tuple of float
            Size of grid as *(n_rows, n_columns)*.
        n0 : float
            Upstream elevation [m].
        nslope : float
            Initial landscape slope [-].
        max_rand : float
            Multiply by slope for max height of a random perturbation.
        dt_day : float
            Time step [days].
        Initial_SL : float
            Initial sea level [m].
        SLRR_m : float
            Sea level rise rate [m / yr].
        SubRate_m: float
            Subsidence rate [m / yr].
        Sub_Start : int
            Row where subsidence starts
        ch_width : float
            Characteristic channel width [m].
        ch_depth : float
            Characteristic channel depth [m].
        ch_discharge : float
            Long-term averaged discharge [m^3 / s].
        A : float
            River-dependent const (1 for meandering; 1.4 for braided).
        c_f : float
            Drag coefficient.
        C_0 : float
            Sediment concentration on bed.
        sed_sg : float
            Sediment specific gravity
        init_cut_frac : float
            Initial cut of the channel into land surface.
        super_ratio : float
            Normalized SE ratio to trigger avulsion.
        short_path : int
            Flag for using shortest path to complete avulsion.
        WL_Z : float
            Elevation that wetlands maintain above SL [m].
        WL_dist : float
            Cell distance beyond channel that wetlands exist.
        blanket_rate_m : float
            "Blanket" deposition rate [m].
        fine_dep_frac : float
            Fraction of channel deposit for adjacent fine deposition.
        splay_type : {0, 1, 2}
            Size/type of splay. Values of *splay_type* correspond to:
            *  0: no splay deposition
            *  1: just the first failed avulsion river cell
            *  2: first failed cell and adjacent cells
        saveavulsions : bool
            Flag for saving avulsion info.
        savecourseupdates : bool
            Flag for saving course updates
        """
        self._params = locals().copy()
        self._params.pop("self")

        np.random.seed(rand_seed)

        # Spatial parameters
        self._dy = spacing[0] * 1000.0
        self._dx = spacing[1] * 1000.0

        n_rows = int(shape[0])
        n_cols = int(shape[1])

        # Initialize elevation grid
        # transverse and longitudinal space
        self._x, self._y = np.meshgrid(
            np.arange(n_cols) * self._dx, np.arange(n_rows) * self._dy
        )
        # eta, elevation
        self._slope = nslope
        self._max_rand = max_rand * nslope
        self._n = n0 - (
            self._slope * self._y + np.random.rand(n_rows, n_cols) * self._max_rand
        )
        self._n -= 0.05

        # self._dn_rc = np.zeros((self._imax))       # change in elevation along river course
        # self._dn_fp = np.zeros_like(self._n)     # change in elevation due to floodplain dep

        self._riv_i = np.zeros(1, dtype=np.int)  # defines first x river locations
        self._riv_j = np.zeros(1, dtype=np.int)  # defines first y river locations
        self._riv_j[0] = self._n.shape[1] / 2

        # Time parameters
        self._dt = dt_day * _SECONDS_PER_DAY  # convert timestep to seconds
        self._time = 0.0

        # Sea level and subsidence parameters
        self._SL = Initial_SL  # starting sea level
        self._SLRR = (
            SLRR_m / _SECONDS_PER_YEAR * self._dt
        )  # sea level rise rate in m (per timestep)
        self._SubRate = (
            SubRate_m / _SECONDS_PER_YEAR * self._dt
        )  # subsidence rate in m (per timestep)
        self._SubStart = Sub_Start  # row where subsidence begins

        # River parameters
        self._nu = (8.0 * (ch_discharge / ch_width) * A * np.sqrt(c_f)) / (
            C_0 * (sed_sg - 1)
        )
        # NEED TO REDO DIFFUSE.PY TO HAVE SIGN OF NU CORRECT (NEG) ABOVE ###
        init_cut = init_cut_frac * ch_depth
        self._super_ratio = super_ratio
        self._short_path = short_path
        self._ch_depth = ch_depth

        # Floodplain and wetland characteristics
        self._WL_Z = WL_Z
        self._WL_dist = WL_dist
        self._blanket_rate = (
            blanket_rate_m / _SECONDS_PER_YEAR
        ) * self._dt  # blanket deposition in m
        self._splay_type = splay_type
        self._frac_fines = fine_dep_frac

        self._sed_flux = 0.0
        self._splay_deposit = np.zeros_like(self._n)

        # Saving information
        self._saveavulsions = saveavulsions
        self._saveupdates = savecourseupdates
        self._save_splay_deposit = False

        if self._saveupdates:
            make_empty_file(self._saveupdates)
        if self._saveavulsions:
            make_empty_file(self._saveavulsions)
        if self._save_splay_deposit:
            make_empty_file(self._savesplay_deposit)

        self._riv_i, self._riv_j = steep_desc.find_course(
            self._n,
            self._riv_i,
            self._riv_j,
            len(self._riv_i),
            self._ch_depth,
            sea_level=self._SL,
        )

        # downcut into new river course by amount determined by init_cut
        downcut.cut_init(self._riv_i, self._riv_j, self._n, init_cut)

        # smooth initial river course elevations using linear diffusion equation
        diffuse.smooth_rc(
            self._dx,
            self._dy,
            self._nu,
            self._dt,
            self._ch_depth,
            self._riv_i,
            self._riv_j,
            self._n,
            self._SL,
            self._slope,
        )

        # initial profile
        self._profile = self._n[self._riv_i, self._riv_j]

        self._profile = self._n[self._riv_i, self._riv_j]

    def to_yaml(self):
        return yaml.dump(self._params)

    @property
    def time(self):
        """Current model time (converted from seconds to days)."""
        return self._time / _SECONDS_PER_DAY

    @property
    def time_step(self):
        """Model time step (converted from seconds to days)."""
        return self._dt / _SECONDS_PER_DAY

    @time_step.setter
    def time_step(self, time_step):
        """Set model time step (time_step is in days)."""
        self._dt = time_step * _SECONDS_PER_DAY

    @property
    def grid_shape(self):
        return self._n.shape

    @property
    def grid_spacing(self):
        return (self._dy, self._dx)

    @property
    def river_x_coordinates(self):
        return self._riv_j * self._dx

    @river_x_coordinates.setter
    def river_x_coordinates(self, new_coords):
        self._riv_j = np.asarray(new_coords) / self._dx

    @property
    def river_y_coordinates(self):
        return self._riv_i * self._dy

    @river_y_coordinates.setter
    def river_y_coordinates(self, new_coords):
        self._riv_i = np.asarray(new_coords) / self._dy

    @property
    def sea_level(self):
        return self._SL

    @property
    def elevation(self):
        return self._n
        # return self._n + self.sea_level

    @elevation.setter
    def elevation(self, new_elev):
        """Set the land surface elevation."""
        self._elevation[:] = new_elev
        # self._elevation[:] = new_elev - self.sea_level

    @property
    def sediment_flux(self):
        return self._sed_flux

    @property
    def avulsions(self):
        return self._avulsion_info

    @property
    def profile(self):
        self._profile[-1] = self._SL - self._ch_depth
        return self._profile

    #    @shoreline.setter
    #    def shoreline(self, shoreline):
    #        self._shoreline = shoreline

    @classmethod
    def from_path(cls, fname):
        """Create a RiverModule object from a file-like object."""
        if fname:
            params = read_params_from_file(fname)
        else:
            params = dict()
        return cls(**params)

    def advance_in_time(self):
        """ Update avulsion model one time step. """
        # if (self._time / _SECONDS_PER_YEAR) > 2000:
        #     self._SLRR = 0.01 / _SECONDS_PER_YEAR * self._dt

        self._riv_i, self._riv_j, self._course_update = steep_desc.update_course(
            self._n,
            self._riv_i,
            self._riv_j,
            self._ch_depth,
            self._slope,
            sea_level=self._SL,
            dx=self._dx,
            dy=self._dy,
        )

        self._n = avulsion_utils.fix_elevations(
            self._n,
            self._riv_i,
            self._riv_j,
            self._ch_depth,
            self._SL,
            self._slope,
            self._dx,
            self._max_rand,
            self._SLRR,
        )

        """ Save every time the course changes? """
        if self._saveupdates and self._course_update > 0:
            with open(self._saveupdates, "a") as file:
                file.write(
                    "%.5f %i \n"
                    % ((self._time / _SECONDS_PER_YEAR), self._course_update)
                )

        """ determine if there is an avulsion & find new path if so """
        (
            (self._riv_i, self._riv_j),
            self._avulsion_type,
            self._loc,
            self._avulse_length,
            self._path_diff,
            self._splay_deposit,
        ) = avulse.find_avulsion(
            self._riv_i,
            self._riv_j,
            self._n,
            self._super_ratio,
            self._SL,
            self._ch_depth,
            self._short_path,
            self._splay_type,
            self._slope,
            self._splay_deposit,
            self._nu,
            self._dt,
            dx=self._dx,
            dy=self._dy,
        )

        """ Save avulsion record. """
        if self._saveavulsions and self._avulsion_type > 0:
            with open(self._saveavulsions, "a") as file:
                file.write(
                    "%.5f %i %i %.5f %.5f\n"
                    % (
                        (self._time / _SECONDS_PER_YEAR),
                        self._avulsion_type,
                        self._loc,
                        self._avulse_length,
                        self._path_diff,
                    )
                )

        """ Save crevasse splay deposits. """
        if self._save_splay_deposit and (self._splay_deposit.sum() > 0):
            np.savetxt(self._save_splay_deposit, self._splay_deposit, "%.8f")

        # need to fill old river channels if coupled to CEM
        if (self._avulsion_type == 1) or (self._avulsion_type == 2):
            self._n = avulsion_utils.fix_elevations(
                self._n,
                self._riv_i,
                self._riv_j,
                self._ch_depth,
                self._SL,
                self._slope,
                self._dx,
                self._max_rand,
                self._SLRR,
            )

        # assert(self._riv_i[-1] != 0)

        """ change elevations according to sea level rise (SLRR)
        (if not coupled -- this occurs in coupling script otherwise) """
        # SLR.elev_change(self._SL, self._n, self._riv_i,
        #                 self._riv_j, self._ch_depth, self._SLRR)

        """ smooth river course elevations using linear diffusion equation """
        self._dn_rc = diffuse.smooth_rc(
            self._dx,
            self._dy,
            self._nu,
            self._dt,
            self._ch_depth,
            self._riv_i,
            self._riv_j,
            self._n,
            self._SL,
            self._slope,
        )

        """ Floodplain sedimentation (use one or the other) """
        # -------------------------------------------------------
        # Deposit blanket across entire subaerial domain: ###
        # FP.dep_blanket(self._SL, self._blanket_rate, self._n,
        #                self._riv_i, self._riv_j, self._ch_depth)

        # Deposit fines adjacent to river channel: ###
        FP.dep_fines(
            self._n, self._riv_i, self._riv_j, self._dn_rc, self._frac_fines, self._SL
        )
        # -------------------------------------------------------

        """ Wetland sedimentation """
        # no wetlands in first version of coupling to CEM ###
        # FP.wetlands(self._SL, self._WL_Z, self._WL_dist * self._dy,
        #             self._n, self._riv_i, self._riv_j, self._x, self._y)

        """ Subsidence """
        subside.linear_subsidence(
            self._n,
            self._riv_i,
            self._riv_j,
            self._ch_depth,
            self._SubRate,
            self._SubStart,
            self._SL,
        )

        """ calculate sediment flux at the river mouth """
        self._sed_flux = flux.calc_qs(
            self._nu,
            self._riv_i,
            self._riv_j,
            self._n,
            self._SL,
            self._ch_depth,
            self._dx,
            self._dy,
            self._dt,
            self._slope,
        )

        self._profile = self._n[self._riv_i, self._riv_j]

        # Update time
        self._time += self._dt
        # update sea level
        self._SL += self._SLRR



#! /usr/bin/env python
"""Basic Model Interface implementation for River Module"""

import numpy as np
import yaml
from bmipy import Bmi

from .rivermodule import RiverModule


class BmiRiverModule(Bmi):

    """The BMI for the River Avulsion Floodplain Evolution Model."""

    _name = "Rafem"
    _input_var_names = (
        "land_surface__elevation",
        "channel_exit__x_coordinate",
        "channel_exit__y_coordinate",
    )
    _output_var_names = (
        "channel_centerline__x_coordinate",
        "channel_centerline__y_coordinate",
        "channel_centerline__elevation",
        "channel_exit_water_sediment~bedload__volume_flow_rate",
        "channel_exit__x_coordinate",
        "channel_exit__y_coordinate",
        "land_surface__elevation",
        "sea_water_surface__elevation",
        # 'avulsion_record',
    )

    def __init__(self):
        """Create a BmiRiver module that is ready for initialization."""
        self._model = None
        self._values = {}
        self._var_units = {}

    def initialize(self, config_file):
        with open(config_file, "r") as fp:
            params = yaml.safe_load(fp)

        params.pop("_version", None)

        self._n_days = params.pop("days", np.finfo("d").max)
        self._model = RiverModule(**params)

        # self._model = RiverModule.from_path(config_file)

        self._values = {
            "channel_centerline__x_coordinate": lambda: self._model.river_x_coordinates,
            "channel_centerline__y_coordinate": lambda: self._model.river_y_coordinates,
            "channel_exit_water_sediment~bedload__volume_flow_rate": lambda: np.array(
                self._model.sediment_flux
            ),
            "channel_exit__x_coordinate": lambda: np.array(
                self._model.river_x_coordinates[-1]
            ),
            "channel_exit__y_coordinate": lambda: np.array(
                self._model.river_y_coordinates[-1]
            ),
            "land_surface__elevation": lambda: np.array(self._model.elevation),
            "channel_centerline__elevation": lambda: self._model.profile,
            "sea_water_surface__elevation": lambda: np.array(self._model.sea_level),
            "avulsion_record": lambda: self._model.avulsions,
        }

        self._var_units = {
            "channel_centerline__x_coordinate": "m",
            "channel_centerline__y_coordinate": "m",
            "channel_exit_water_sediment~bedload__volume_flow_rate": "m^3 s^-1",
            "channel_exit__x_coordinate": "m",
            "channel_exit__y_coordinate": "m",
            "land_surface__elevation": "m",
            "channel_centerline__elevation": "m",
            "sea_water_surface__elevation": "m",
            "avulsion_record": "none",
        }

        self._var_type = {}
        for name in self._input_var_names + self._output_var_names:
            self._var_type[name] = str(np.dtype(float))

        self._var_grid = {
            "channel_centerline__x_coordinate": 1,
            "channel_centerline__y_coordinate": 1,
            "channel_exit_water_sediment~bedload__volume_flow_rate": 2,
            "channel_exit__x_coordinate": 2,
            "channel_exit__y_coordinate": 2,
            "land_surface__elevation": 0,
            "channel_centerline__elevation": 1,
            "sea_water_surface__elevation": 2,
            "avulsion_record": None,
        }

        self._grid_rank = {0: 2, 1: 1, 2: 0}

    def update(self):
        self._model.advance_in_time()
        self.river_mouth_location = (
            self._model.river_x_coordinates[-1],
            self._model.river_y_coordinates[-1],
        )

    def update_frac(self, time_frac):
        """Update model by a fraction of a time step."""
        time_step = self.get_time_step()
        self._model.time_step = time_frac * time_step
        self.update()
        self._model.time_step = time_step

    def update_until(self, then):
        """Update model until a particular time."""
        n_steps = (then - self.get_current_time()) / self.get_time_step()

        for _ in range(int(n_steps)):
            self.update()
        # self.update_frac(n_steps - int(n_steps))

    def finalize(self):
        pass

    def get_var_type(self, name):
        return self._var_type[name]

    def get_var_units(self, name):
        return self._var_units[name]

    def get_var_nbytes(self, name):
        return self._values[name]().nbytes

    def get_var_itemsize(self, name):
        return np.dtype(self.get_var_type(name)).itemsize

    def get_var_location(self, name):
        return "node"

    def get_var_grid(self, name):
        return self._var_grid[name]

    def get_grid_rank(self, grid):
        return self._grid_rank[grid]

    def get_grid_size(self, grid):
        if grid == 0:
            return int(np.prod(self._model.grid_shape))
        elif grid == 1:
            return len(self._model.river_x_coordinates)
        elif grid == 2:
            return 1
        else:
            raise KeyError(grid)

    def get_grid_shape(self, grid, shape):
        if grid == 0:
            shape[:] = self._model.grid_shape
        elif grid == 1:
            shape[:] = self._model.river_x_coordinates.shape
        elif grid == 2:
            shape[0] = 1  # = (1, )
        else:
            raise KeyError(grid)

        return shape

        # if out is None:
        #     return shape
        # else:
        #     out[:] = shape
        #     return out

    def get_grid_spacing(self, grid, spacing):
        if grid == 0:
            spacing[:] = self._model.grid_spacing
        else:
            raise KeyError(grid)
        return spacing

        # if out is None:
        #     return self._model.grid_spacing
        # else:
        #     out[:] = self._model.grid_spacing
        #     return out

    def get_grid_origin(self, grid, origin):
        if grid == 0:
            origin[:] = (0.0, 0.0)
        else:
            raise KeyError(grid)
        return origin
        # if out is None:
        #     return origin
        # else:
        #     out[:] = origin
        #     return out

    def get_grid_type(self, grid):
        if grid == 0:
            return "uniform_rectilinear"
        elif grid == 1:
            return "vector"
        elif grid == 2:
            return "scalar"
        else:
            raise KeyError(grid)

    def get_value(self, name, dest):
        dest[:] = self._values[name]().flat
        return dest

    def set_value(self, var_name, new_vals):
        """Set model values."""
        if var_name not in self._input_var_names:
            raise KeyError(var_name)

        if var_name == "land_surface__elevation":
            np.copyto(self._model.elevation.reshape((-1,)), new_vals.reshape((-1,)))
        elif var_name == "channel_exit__x_coordinate":
            self._model.river_x_coordinates = np.append(
                self._model.river_x_coordinates, new_vals
            )
        elif var_name == "channel_exit__y_coordinate":
            self._model.river_y_coordinates = np.append(
                self._model.river_y_coordinates, new_vals
            )

        # Remove duplicate river mouth coordinates (if they exist).
        # This seems clunky... must be better way to get values without
        # duplicating each time?
        # if (self._model.river_x_coordinates[-1] == self._model.river_x_coordinates[-2] and
        #    self._model.river_y_coordinates[-1] == self._model.river_y_coordinates[-2]):
        #    self._model.river_x_coordinates.pop()
        #    self._model.river_y_coordinates.pop()

    def get_component_name(self):
        return self._name

    def get_input_var_names(self):
        return self._input_var_names

    def get_output_var_names(self):
        return self._output_var_names

    def get_input_item_count(self):
        return len(self._input_var_names)

    def get_output_item_count(self):
        return len(self._output_var_names)

    def get_start_time(self):
        return 0.0

    def get_end_time(self):
        return self._n_days

    def get_current_time(self):
        return self._model.time

    def get_time_step(self):
        return self._model.time_step

    def get_time_units(self):
        return "d"

    def get_grid_node_count(self, grid):
        raise NotImplementedError("get_grid_node_count")

    def get_grid_edge_count(self, grid):
        raise NotImplementedError("get_grid_edge_count")

    def get_grid_face_count(self, grid):
        raise NotImplementedError("get_grid_face_count")

    def get_grid_edge_nodes(self, grid, edge_nodes):
        raise NotImplementedError("get_grid_edge_nodes")

    def get_grid_face_edges(self, grid, face_edges):
        raise NotImplementedError("get_grid_face_edges")

    def get_grid_face_nodes(self, grid, face_nodes):
        raise NotImplementedError("get_grid_edge_nodes")

    def get_grid_nodes_per_face(self, grid, nodes_per_face):
        raise NotImplementedError("get_grid_nodes_per_face")

    def get_grid_x(self, grid, x):
        raise NotImplementedError("get_grid_x")

    def get_grid_y(self, grid, y):
        raise NotImplementedError("get_grid_y")

    def get_grid_z(self, grid, z):
        raise NotImplementedError("get_grid_z")

    def get_value_at_indices(self, name, dest, inds):
        raise NotImplementedError("get_value_at_indices")

    def get_value_ptr(self, name):
        raise NotImplementedError("get_value_ptr")

    def set_value_at_indices(self, name, ids, src):
        raise NotImplementedError("set_value_at_indices")



#! /usr/bin/env python
from setuptools import find_packages, setup

import versioneer


setup(
    name="rafem",
    version=versioneer.get_version(),
    author="Katherine Ratliff",
    author_email="k.ratliff@duke.edu",
    description="River Avulsion Flooplain Evolution Model",
    long_description=open("README.rst", encoding="utf-8").read(),
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    url="https://github.com/katmratliff/avulsion-bmi",
    license="MIT",
    install_requires=open("requirements.txt", "r").read().splitlines(),
    packages=find_packages(),
    entry_points={"console_scripts": ["rafem=rafem.main:rafem"]},
    cmdclass=versioneer.get_cmdclass(),
)



#! /usr/local/bin/python
import numpy as np

from . import FP, downcut, steep_desc
from .avulsion_utils import (
    channel_is_superelevated,
    fill_abandoned_channel,
    find_path_length,
    find_point_in_path,
    find_riv_path_length,
)
from .diffuse import calc_crevasse_dep


def avulse_to_new_path(
    z, old, new, sea_level, channel_depth, avulsion_type, slope, dx=1.0, dy=1.0,
):
    """Avulse the river to a new path.

    Given two river paths, *old* and *new*, avulse the river to a new river
    path. If the end point of the new path is contained in the old river
    path, the resulting path is the new path up until this point and then
    the old path. Otherwise, the resulting path is the new river path and
    will be downcut.

    Parameters
    ----------
    z : ndarray
        2D array of elevations.
    old : tuple of array_like
        Tuple of i and j indices (into *z*) for the old path.
    new : tuple of array_like
        Tuple of i and j indices (into *z*) for the new path.
    sea_level : float
        Elevation of sea level.
    channel_depth : float
        Depth of the channel.
    avulsion_type : {0, 1, 2, 3}
        The type of the avulsion.
    dx : float, optional
        Spacing of columns of *z*.
    dy : float, optional
        Spacing of rows of *z*.

    Returns
    -------
    tuple
        Tuple of the new river path (as i, j indices) and the, possibly
        changed, avulsion type.

    Examples
    --------
    The following example uses a grid that looks like::

        o  +  *  *
        *  o  +  *
        *  *  +  *
        *  *  o  *
        *  o  *  *

    The old path is marked by `o`, the new path but `+`. The paths overlap
    (2, 2).

    >>> import numpy as np
    >>> z = np.ones((5, 4), dtype=float)

    >>> old = np.array((0, 1, 2, 3, 4)), np.array((0, 1, 2, 2, 1))
    >>> new = np.array((0, 1, 2)), np.array((1, 2, 2))
    >>> (new, atype) = avulse_to_new_path(z, old, new, 0., 0., 0)

    The new path follows the new path until the common point and then
    follows the old path. The new avulsion type is now 2.

    >>> new
    (array([0, 1, 2, 3, 4]), array([1, 2, 2, 2, 1]))
    >>> atype
    2

    In this example the old and new paths do not overlap::

        o  +  *  *
        *  o  +  *
        *  *  o  +
        *  *  o  +
        *  o  *  +

    >>> old = np.array((0, 1, 2, 3, 4)), np.array((0, 1, 2, 2, 1))
    >>> new = np.array((0, 1, 2, 3, 4)), np.array((1, 2, 3, 3, 3))
    >>> (new, atype) = avulse_to_new_path(z, old, new, 0., 0., 0)

    The new path is now, in fact, the actual new path and the avulsion
    type is unchanged.

    >>> new
    (array([0, 1, 2, 3, 4]), array([1, 2, 3, 3, 3]))
    >>> atype
    0
    """
    old_i, old_j = old
    new_i, new_j = new
    # sets avulsion to be regional, may be updated again below (if local)

    # maybe this should be len(test_old_x)-1?
    ind = find_point_in_path((old_i, old_j), (new_i[-1], new_j[-1]))

    if ind is not None:
        avulsion_type = 2

        downcut.cut_local(new_i, new_j, z, dx=dx, dy=dy)

        new_i = np.append(new_i, old_i[ind + 1 :])
        new_j = np.append(new_j, old_j[ind + 1 :])
    else:
        max_cell_h = slope * dx
        if (z[new_i[-1], new_j[-1]] - sea_level) < (0.001 * max_cell_h):
            z[new_i[-1], new_j[-1]] = (0.001 * max_cell_h) + sea_level

        downcut.cut_new(new_i, new_j, z, sea_level, channel_depth, dx=dx, dy=dy)

    return (new_i, new_j), avulsion_type


# determines if there is an avulsion along river course
def find_avulsion(
    riv_i,
    riv_j,
    n,
    super_ratio,
    current_SL,
    ch_depth,
    short_path,
    splay_type,
    slope,
    splay_depth,
    nu,
    dt,
    dx=1.0,
    dy=1.0,
):
    new = riv_i, riv_j
    old = riv_i, riv_j
    avulsion_type = 0
    a = 0
    loc = 0
    avulse_length = 0
    new_length = 0
    avul_locs = np.zeros(0, dtype=np.int)
    path_slopes = np.zeros(0)
    crevasse_locs = np.zeros(3, dtype=np.int)
    path_diff = np.zeros(0)
    path_difference = 0

    old_length = find_riv_path_length(n, old, current_SL, ch_depth, slope, dx=dx, dy=dy)

    for a in range(1, len(riv_i) - 1):
        if channel_is_superelevated(
            n,
            (riv_i[a], riv_j[a]),
            (riv_i[a - 1], riv_j[a - 1]),
            ch_depth,
            super_ratio,
            current_SL,
        ):

            # if superelevation greater than trigger ratio, determine
            # new steepest descent path
            new = steep_desc.find_course(
                n, riv_i, riv_j, a, ch_depth, sea_level=current_SL
            )

            if n[new[0][-1], new[1][-1]] < current_SL:
                new_length = find_riv_path_length(
                    n, new, current_SL, ch_depth, slope, dx=dx, dy=dy
                )
            else:
                new_length = find_path_length(
                    n, new, current_SL, ch_depth, slope, dx=dx, dy=dy
                )

            if new_length < old_length:
                # calculate slope of new path

                if len(new[0][a:]) <= 1:
                    avulsed_length = find_path_length(
                        n,
                        (new[0][a - 1 :], new[1][a - 1 :]),
                        current_SL,
                        ch_depth,
                        slope,
                        dx=dx,
                        dy=dy,
                    )
                    slope_new_path = (
                        n[new[0][-2], new[1][-2]] - n[new[0][-1], new[1][-1]]
                    ) / avulsed_length

                elif n[new[0][-1], new[1][-1]] < current_SL:
                    avulsed_length = find_riv_path_length(
                        n,
                        (new[0][a:], new[1][a:]),
                        current_SL,
                        ch_depth,
                        slope,
                        dx=dx,
                        dy=dy,
                    )
                    slope_new_path = (
                        n[new[0][a], new[1][a]] - n[new[0][-1], new[1][-1]]
                    ) / avulsed_length

                else:
                    avulsed_length = find_path_length(
                        n,
                        (new[0][a:], new[1][a:]),
                        current_SL,
                        ch_depth,
                        slope,
                        dx=dx,
                        dy=dy,
                    )
                    slope_new_path = (
                        n[new[0][a], new[1][a]] - n[new[0][-1], new[1][-1]]
                    ) / avulsed_length

                avul_locs = np.append(avul_locs, a)
                path_slopes = np.append(path_slopes, slope_new_path)
                path_diff = np.append(path_diff, (old_length - new_length))

            crevasse_locs = np.vstack((crevasse_locs, [new[0][a], new[1][a], a]))

    if crevasse_locs.sum() > 0:
        crevasse_locs = np.delete(crevasse_locs, 0, 0)

    if avul_locs.size > 0:

        max_slope = np.argmax(path_slopes)
        loc = avul_locs[max_slope]
        path_difference = path_diff[max_slope]

        new = steep_desc.find_course(
            n, riv_i, riv_j, loc, ch_depth, sea_level=current_SL
        )

        avulsion_type = 1

        new, avulsion_type = avulse_to_new_path(
            n,
            (riv_i[loc - 1 :], riv_j[loc - 1 :]),
            (new[0][loc - 1 :], new[1][loc - 1 :]),
            current_SL,
            ch_depth,
            avulsion_type,
            slope,
            dx=dx,
            dy=dy,
        )

        new = (np.append(riv_i[: loc - 1], new[0]), np.append(riv_j[: loc - 1], new[1]))

        avulse_length = find_riv_path_length(
            n, (riv_i[loc:], riv_j[loc:]), current_SL, ch_depth, slope, dx=dx, dy=dy
        )

        # fill up old channel... could be some fraction in the future
        # (determines whether channels are repellors or attractors)
        fill_abandoned_channel(
            loc, n, new, riv_i, riv_j, current_SL, ch_depth, slope, dx
        )

        crevasse_locs = np.delete(crevasse_locs, max_slope, 0)

    else:
        new = riv_i, riv_j

    if (crevasse_locs.sum() > 0) and (splay_type > 0):

        n_before_splay = np.copy(n)

        # Don' think we need to worry about preserving old river elevations??
        # old_river_elevations = n[riv_i, riv_j]
        new_river_elevations = n[new[0], new[1]]

        for i in range(crevasse_locs.shape[0]):

            splay_dep = calc_crevasse_dep(
                dx,
                dy,
                nu,
                dt,
                ch_depth,
                riv_i,
                riv_j,
                n,
                current_SL,
                slope,
                crevasse_locs[i][2],
            )

            if splay_dep > 0:
                FP.dep_splay(
                    n,
                    (crevasse_locs[i][0], crevasse_locs[i][1]),
                    splay_dep,
                    splay_type=splay_type,
                )

        # n[riv_i, riv_j] = old_river_elevations
        n[new[0], new[1]] = new_river_elevations
        n_splay = n - n_before_splay
        splay_depth += n_splay

    return (new, avulsion_type, loc, avulse_length, path_difference, splay_depth)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import os
import pathlib
import re
import sys
from collections import OrderedDict
from functools import partial

import click
import matplotlib.pyplot as plt
import numpy as np
import yaml

from ._version import get_versions
from .riverbmi import BmiRiverModule

__version__ = get_versions()["version"]

out = partial(click.secho, bold=True, err=True)
err = partial(click.secho, fg="red", err=True)


def setup_yaml_with_canonical_dict():
    """ https://stackoverflow.com/a/8661021 """
    yaml.add_representer(
        OrderedDict,
        lambda self, data: self.represent_mapping(
            "tag:yaml.org,2002:map", data.items()
        ),
        Dumper=yaml.SafeDumper,
    )

    def repr_ordered_dict(self, data):
        return self.represent_mapping("tag:yaml.org,2002:map", data.items())

    yaml.add_representer(dict, repr_ordered_dict, Dumper=yaml.SafeDumper)

    def repr_dict(self, data):
        return self.represent_mapping(
            "tag:yaml.org,2002:map", sorted(data.items(), key=lambda t: t[0])
        )

    yaml.add_representer(dict, repr_dict, Dumper=yaml.SafeDumper)

    # https://stackoverflow.com/a/45004464
    def repr_str(dumper, data):
        if "\n" in data:
            return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
        return dumper.represent_str(data)

    yaml.add_representer(str, repr_str, Dumper=yaml.SafeDumper)

    def repr_tuple(dumper, data):
        return dumper.represent_sequence("tag:yaml.org,2002:seq", list(data))

    yaml.add_representer(tuple, repr_tuple, Dumper=yaml.SafeDumper)

    yaml.add_implicit_resolver(
        "tag:yaml.org,2002:float",
        re.compile(
            r"""^(?:
         [-+]?(?:[0-9][0-9_]*)\.[0-9_]*(?:[eE][-+]?[0-9]+)?
        |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
        |[-+]?\.[0-9_]+(?:[eE][-+]?[0-9]+)
        |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\.[0-9_]*
        |[-+]?\.(?:inf|Inf|INF)
        |\.(?:nan|NaN|NAN))$""",
            re.X,
        ),
        list("-+0123456789."),
        Loader=yaml.SafeLoader,
    )


setup_yaml_with_canonical_dict()


def empty_bmi_var_array(bmi, name):
    return np.empty(bmi.get_var_nbytes(name), dtype=np.uint8).view(
        dtype=bmi.get_var_type(name)
    )


class RafemOutputWriter:
    def __init__(self, bmi, output_interval=1):
        self._bmi = bmi
        # self._run_id = run_id
        self._output_interval = output_interval
        self._steps = 0
        self._prefix = pathlib.Path("output")

        self._outputs = {
            "elevation": self._prefix / "elevation",
            "river": self._prefix / "river",
            "profile": self._prefix / "profile",
        }

        self._make_dirs()
        self._make_buffers()

    def _make_dirs(self):
        os.mkdir(self._prefix)
        for dir_ in self._outputs.values():
            os.mkdir(dir_)

    def _make_buffers(self):
        self._z = empty_bmi_var_array(self._bmi, "land_surface__elevation")
        self._x = empty_bmi_var_array(self._bmi, "channel_centerline__x_coordinate")
        self._y = empty_bmi_var_array(self._bmi, "channel_centerline__y_coordinate")
        self._prof = empty_bmi_var_array(self._bmi, "channel_centerline__elevation")

    def _update_buffers(self):
        self._bmi.get_value("land_surface__elevation", self._z)
        self._bmi.get_value("channel_centerline__x_coordinate", self._x)
        self._bmi.get_value("channel_centerline__y_coordinate", self._y)
        self._bmi.get_value("channel_centerline__elevation", self._prof)

    def time_stamp(self):
        return str(self._steps)

    def _save_output(self):
        self._update_buffers()

        np.savetxt(
            self._outputs["elevation"] / "elev_{0}.out".format(self.time_stamp()),
            self._z,
            fmt="%.5f",
            header=os.linesep.join(
                ["version: {0}".format(__version__), "Elevation [m]"]
            ),
        )
        np.savetxt(
            self._outputs["river"] / "riv_{0}.out".format(self.time_stamp()),
            list(zip(self._x, self._y)),
            fmt="%.5f",
            header=os.linesep.join(
                ["version: {0}".format(__version__), "X [m], Y [m]"]
            ),
        )
        np.savetxt(
            self._outputs["profile"] / "prof_{0}.out".format(self.time_stamp()),
            self._prof,
            fmt="%.5f",
            header=os.linesep.join(
                ["version: {0}".format(__version__), "Elevation [m]"]
            ),
        )

    @property
    def prefix(self):
        return str(self._prefix)

    def update(self, n_steps):
        self._steps += n_steps
        if self._steps % self._output_interval == 0:
            self._save_output()


@click.group(chain=True)
@click.version_option()
@click.option(
    "--cd",
    default=".",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, readable=True),
    help="chage to directory, then execute",
)
def rafem(cd) -> None:
    """The River Avulsion and Floodplain Evolution Model (RAFEM)

    \b
    Examples:

      Create a folder with example input files,

        $ mkdir example && rafem setup

      Run a simulation using the examples input files,

        $ cd example && rafem run

      Commands can also be chained together. The folling will setup
      a simulation, run it, and then plot elevations.

        $ mkdir example && rafem setup run plot elevation
    """
    os.chdir(cd)


@rafem.command()
@click.option("-v", "--verbose", is_flag=True, help="Emit status messages to stderr.")
@click.option("--dry-run", is_flag=True, help="Do not actually run the model")
def run(dry_run: bool, verbose: bool) -> None:
    """Run a simulation."""
    run_dir = pathlib.Path.cwd()
    config_file = run_dir / "rafem.yaml"

    message = []
    if not config_file.is_file():
        message.append("missing RAFEM configuration file: {0}".format(config_file))
    if (run_dir / "output").exists():
        message.append(
            "RAFEM output directory already exists: {0}".format(run_dir / "output")
        )
    if message:
        err(os.linesep.join(message))
        raise click.Abort(os.linesep.join(message))

    if dry_run:
        out("Nothing to do. 😴")
    else:
        avulsion = BmiRiverModule()
        avulsion.initialize("rafem.yaml")

        with open("rafem.yaml", "r") as fp:
            params = yaml.safe_load(fp)

        spacing = 10
        n_days = params["days"]  # + years * 365
        n_steps = int(n_days / avulsion.get_time_step())

        if dry_run:
            out(avulsion._model.to_yaml())
        elif n_steps == 0:
            out("Nothing to do (years == 0). 😴")
        else:
            output = RafemOutputWriter(avulsion, output_interval=spacing)

            with click.progressbar(
                range(n_steps),
                label=" ".join(["🚀", str(run_dir)]),
                item_show_func=lambda step: "day {0} of {1}".format(
                    int(0 if step is None else step * avulsion.get_time_step()), n_days
                ),
            ) as bar:
                for step in bar:
                    avulsion.update()
                    output.update(1)

            out("💥 Finished! 💥")
            out("Output written to {0}".format(output.prefix))

        avulsion.finalize()


@rafem.command()
@click.argument("infile", type=click.Choice(["rafem"]))
def show(infile: str) -> None:
    """Show example input files."""
    print(_contents_of_input_file(infile))


@rafem.command()
def setup() -> None:
    """Setup a folder of input files for a simulation."""
    files = [pathlib.Path("rafem.yaml")]

    existing_files = [name for name in files if name.exists()]
    if existing_files:
        for name in existing_files:
            err(
                f"{name}: File exists. Either remove and then rerun or setup in a different folder",
            )
    else:
        for fname in files:
            with open(fname, "w") as fp:
                print(_contents_of_input_file(fname.stem), file=fp)

    if existing_files:
        raise click.Abort()


def _contents_of_input_file(infile: str) -> str:

    # Avulsion module parameters
    conf = OrderedDict(
        (
            ("_version", __version__),
            # Space
            ("shape", [120, 200]),  # Length x Width (km)
            ("spacing", [0.1, 0.1]),  # dy, dx (km)
            ("n0", 5.0),  # upstream elevation
            ("nslope", 0.001),  # initial landscape slope
            (
                "max_rand",
                0.1,
            ),  # multiply by slope for max height of a random perturbation
            # Time
            ("days", 7),  # number of days to run for
            ("dt_day", 0.01),  # timestep (days)
            # Random seed
            ("rand_seed", 623),  # seed for random number generator
            # Sea level and subsidence parameters
            ("Initial_SL", 0.0),  # initial sea level
            ("SLRR_m", 0.0),  # sea level rise rate (m/yr)
            ("SubRate_m", 0.0),  # subsidence rate (m/yr)
            ("Sub_Start", 0),  # row where subsidence starts
            # River characteristics
            ("ch_width", 10.0),  # characteristic channel width (m)
            ("ch_depth", 1.0),  # characteristic channel depth (m)
            ("ch_discharge", 10.0),  # long-term averaged discharge (m^3/s)
            ("A", 1.0),  # river-dependent const. (1 for meandering; 1.4 for braided)
            ("c_f", 0.01),  # drag coefficient
            ("C_0", 1.0),  # sediment concentration on bed
            ("sed_sg", 2.65),  # sediment specific gravity
            ("init_cut_frac", 1),  # initial cut of the channel into land surface
            # Avulsion parameters
            ("super_ratio", 1.0),  # normalized SE ratio to trigger avulsion
            ("short_path", 1),  # flag for using shortest path to complete avulsion
            # Floodplain and Wetland parameters
            ("WL_Z", 0.0),  # elevation that wetlands maintain above SL
            ("WL_dist", 0),  # cell distance beyond channel that wetlands exist
            ("blanket_rate_m", 0.0),  # "blanket" deposition rate (m)
            (
                "fine_dep_frac",
                0.0,
            ),  # fraction of channel deposit for adjacent fine deposition
            ("splay_type", 2),  # size/type of splay
            # Splay types:
            # splay_type = 0: no splay deposition
            # splay_type = 1: just the first failed avulsion river cell
            # splay_type = 2: first failed cell and adjacent cells
            # Saving information
            ("saveavulsions", False),  # flag for saving avulsion info
            ("savecourseupdates", False),  # flag for saving course updates
        )
    )

    contents = {
        "rafem": yaml.safe_dump(conf, default_flow_style=False),
    }

    return contents[infile]


@rafem.command()
@click.option("--time", default=-1, type=int, help="time step to plot")
@click.argument("value", type=click.Choice(["elevation", "profile"]))
def plot(value: str, time: int) -> None:
    """Plot output from a simulation."""
    basepath = pathlib.Path("output") / value

    if time == -1:

        def _alphanum_key(key):
            def _convert(text):
                return int(text) if text.isdigit() else text

            return [_convert(c) for c in re.split("([0-9]+)", key)]

        files = sorted(
            [name.name for name in basepath.glob("*.out")], key=_alphanum_key
        )
        filepath = basepath / files[-1]
    else:
        filepath = list(basepath.glob("*_{0}.out".format(time)))[0]
    out(str(filepath))

    with open("rafem.yaml", "r") as fp:
        params = yaml.safe_load(fp)
    shape = params["shape"]

    z = np.loadtxt(filepath, delimiter=",", comments="#")
    if value == "elevation":
        plt.imshow(z.reshape(shape), origin="lower", cmap="terrain")
        plt.colorbar().ax.set_label("Elevation (m)")
    elif value == "profile":
        plt.plot(z)

    plt.show()



#! /usr/local/bin/python
import numpy as np

from . import downcut
from .avulsion_utils import lowest_cell_elev, sort_lowest_neighbors


def lowest_neighbor(n, sub):
    """Find lowest neighbor value around a point.
    Parameters
    ----------
    n : ndarray
        Grid of elevations.
    sub : tuple of int
        Row/column subscript into elevation matrix.
    Returns
    -------
    tuple of int
        Subscripts into elevation matrix of the lowest neighbor.
    """
    i, j = sub

    if j == n.shape[1] - 1:
        di, dj = np.array([0, 1, 1]), np.array([-1, -1, 0])
    elif j == 0:
        di, dj = np.array([1, 1, 0]), np.array([0, 1, 1])
    else:
        di, dj = np.array([0, 1, 1, 1, 0]), np.array([-1, -1, 0, 1, 1])

    lowest = np.argmin(n[i + di, j + dj])

    return i + di[lowest], j + dj[lowest]


def lowest_neighbor_prograde(n, sub):
    i, j = sub

    if j == n.shape[1] - 1:
        di, dj = np.array([0, 1, 1]), np.array([-1, -1, 0])
    elif j == 0:
        di, dj = np.array([1, 1, 0]), np.array([0, 1, 1])
    else:
        di, dj = np.array([0, 1, 1, 1, 0]), np.array([-1, -1, 0, 1, 1])

    subaerial_low = min(x for x in (n[i + di, j + dj]) if x > 0)
    lowest = np.where((n[i + di, j + dj]) == subaerial_low)[0][0]

    return i + di[lowest], j + dj[lowest]


def below_sea_level(z, sea_level):
    """Check if an elevation is below sea level.
    Parameters
    ----------
    z : float
        Elevation.
    sea_level : float
        Elevation of sea level.
    Returns
    -------
    boolean
        `True` if at or below sea level. Otherwise, `False`.
    """
    return z <= sea_level


def at_river_mouth(z, sub, z0):
    """Check if a cell is at the river mouth.
    Parameters
    ----------
    z : ndarray
        2D-array of elevations.
    sub : tuple of int
        Row and column subscript into *z*.
    z0 : float
        Elevation of sea level (or `None`).
    Returns
    -------
    boolean
        True if the cell at the given subscript is at the river mouth.
    """
    try:
        return sub[0] == z.shape[0] - 1 or below_sea_level(z[sub], z0)
    except IndexError:
        return True


def at_end_of_domain(z, sub):
    """Check if a cell a river mouth at the end of domain.
    Parameters
    ----------
    z : ndarray
        2D-array of elevations.
    sub : tuple of int
        Row and column subscript into *z*.
    Returns
    -------
    boolean
        True if the cell at the given subscript is at the river mouth.
    """
    try:
        return sub[0] == z.shape[0] - 1
    except IndexError:
        return True


def riv_cell_at_sea_level(z, sub, z0):
    """Check if a river cell is at sea level.
    Parameters
    ----------
    z : ndarray
        2D-array of elevations.
    sub : tuple of int
        Row and column subscript into *z*.
    z0 : float
        Elevation of sea level (or `None`).
    Returns
    -------
    boolean
        True if the cell at the given subscript is at the river mouth.
    """
    try:
        below_sea_level(z[sub], z0)
    except IndexError:
        return True


def find_course(z, riv_i, riv_j, SE_loc, channel_depth, sea_level=None):
    """Find the course of a river.
    Given a river course as subscripts, (*i*, *j*), into an array of
    elevations, (*z*), find the river path until the coast (*sea_level*) or
    the end of the domain.
    Parameters
    ----------
    z : ndarray
        Grid elevations.
    riv_i : array_like of int
        Row subscripts (into *n*) for river.
    riv_j : array_like of int
        Column subscripts (into *n*) for river.
    sea_level : None, optional
        The current sea level.
    Returns
    -------
    tuple of array_like
        Row and column indices of the new river path.
    Examples
    --------
    >>> import numpy as np
    >>> z = np.array([[4., 3., 4.],
    ...               [2., 3., 3.],
    ...               [2., 1., 2.]])
    >>> riv_i, riv_j = np.zeros(9, dtype=int), np.zeros(9, dtype=int)
    >>> riv_i[0], riv_j[0] = 0, 1
    >>> find_course(z, riv_i[:1], riv_j[:1], 1)
    (array([0, 1, 2]), array([1, 0, 1]))
    >>> find_course(z, riv_i[:1], riv_j[:1], sea_level=2.)
    (array([0, 1]), array([1, 0]))
    >>> z = np.array([[4., 3., 4.],
    ...               [2., 3., 3.],
    ...               [2., 1., 2.],
    ...               [2., 1.5, 2]])
    >>> find_course(z, riv_i[:1], riv_j[:1], sea_level=0.)
    (array([0, 1, 2, 3]), array([1, 0, 1, 1]))
    >>> z
    """
    # function to find the steepest descent route
    # note: this needs to be improved to remove potential bias that may occur
    # if two or more cells have the steepest descent elevation
    old_course = zip(riv_i, riv_j)

    n_levee = np.copy(z)
    n_levee[riv_i, riv_j] += channel_depth

    riv_i = riv_i[:SE_loc]
    riv_j = riv_j[:SE_loc]

    assert riv_i.size > 0 and riv_j.size > 0

    if sea_level is None:
        sea_level = -np.finfo(float).max

    for n in range(1, riv_i.size):
        if at_end_of_domain(z, (riv_i[n - 1], riv_j[n - 1])):
            return riv_i[:n], riv_j[:n]

    # for n in range(1, riv_i.size):
    #     if riv_cell_at_sea_level(z, (riv_i[n - 1], riv_j[n - 1]), sea_level):
    #         return riv_i[:n-1], riv_j[:n-1]

    new_i = np.empty(z.size, dtype=np.int)
    new_j = np.empty(z.size, dtype=np.int)

    new_i[: len(riv_j)] = riv_i[:]
    new_j[: len(riv_i)] = riv_j[:]

    pits = True
    while pits:
        for n in range(riv_i.size, new_i.size):
            # if at_river_mouth(z, (new_i[n - 1], new_j[n - 1]), sea_level):
            #     pits = False
            #     break

            if at_end_of_domain(z, (new_i[n - 1], new_j[n - 1])):
                pits = False
                break

            sorted_n = sort_lowest_neighbors(n_levee, (new_i[n - 1], new_j[n - 1]))

            if (sorted_n[0][0], sorted_n[1][0]) not in zip(
                new_i[: n - 1], new_j[: n - 1]
            ):
                downstream_ij = (sorted_n[0][0], sorted_n[1][0])
            elif (sorted_n[0][1], sorted_n[1][1]) not in zip(
                new_i[: n - 1], new_j[: n - 1]
            ):
                downstream_ij = (sorted_n[0][1], sorted_n[1][1])
            elif (sorted_n[0][2], sorted_n[1][2]) not in zip(
                new_i[: n - 1], new_j[: n - 1]
            ):
                downstream_ij = (sorted_n[0][2], sorted_n[1][2])
            else:
                raise RuntimeError("river course is going crazy!")

            if downstream_ij not in old_course and below_sea_level(
                z[downstream_ij], sea_level
            ):
                pits = False
                break

            if downstream_ij in old_course:
                new_i[n], new_j[n] = downstream_ij
                n += 1
                pits = False
                break

            if z[downstream_ij] > z[new_i[n - 1], new_j[n - 1]]:
                new_i[n], new_j[n] = downstream_ij
                z[new_i[n - 1], new_j[n - 1]] += 1e-6
            else:
                new_i[n], new_j[n] = downstream_ij

            # new_i[n], new_j[n] = lowest_neighbor(z, (new_i[n - 1], new_j[n - 1]))

    if n == 0:
        raise RuntimeError("new river length is zero!")

    return new_i[:n], new_j[:n]


def update_course(z, riv_i, riv_j, ch_depth, slope, sea_level=None, dx=1.0, dy=1.0):

    if sea_level is None:
        sea_level = -np.finfo(float).max

    course_update = 0

    last_elev = z[riv_i[-1], riv_j[-1]] + ch_depth - sea_level
    max_cell_h = slope * dx

    test_elev = np.copy(z)
    test_elev -= sea_level
    test_elev[riv_i, riv_j] += 2 * ch_depth

    low_adj_cell = lowest_cell_elev(test_elev, (riv_i[-1], riv_j[-1]))

    # check for coastal avulsion (happens if river is prograding too far alongshore)
    if (
        (
            riv_i[-1] == riv_i[-2] == riv_i[-3] == riv_i[-4] == riv_i[-5]
        )  # if last 5 river cells flowing alongshore
        and (z[riv_i[-2] + 1, riv_j[-2]] > sea_level)
        # (z[riv_i[-1+1],riv_j[-1]] > sea_level) and  # if land between river and ocean for last 3 cells
        and (z[riv_i[-3] + 1, riv_j[-3]] > sea_level)
        and (z[riv_i[-4] + 1, riv_j[-4]] > sea_level)
        and (z[riv_i[-5] + 1, riv_j[-5]] > sea_level)
        and (z[riv_i[-1] + 2, riv_j[-1]] <= sea_level)
        and (z[riv_i[-2] + 2, riv_j[-2]] <= sea_level)
        and (z[riv_i[-3] + 2, riv_j[-3]] <= sea_level)
        and (z[riv_i[-4] + 2, riv_j[-4]] <= sea_level)
        and (z[riv_i[-5] + 2, riv_j[-5]] <= sea_level)
    ):

        # fill up old river mouth
        z[riv_i[-1], riv_j[-1]] += ch_depth

        # turn river towards ocean
        riv_i[-1] = riv_i[-2] + 1
        riv_j[-1] = riv_j[-2]

        if (z[riv_i[-1], riv_j[-1]] - sea_level) < (0.001 * max_cell_h):
            z[riv_i[-1], riv_j[-1]] = (0.001 * max_cell_h) + sea_level

        # z[riv_i[-1],riv_j[-1]] -= ch_depth

        downcut.cut_new(riv_i[-3:], riv_j[-3:], z, sea_level, ch_depth, dx=dx, dy=dy)

        course_update = 7  # coastal avulsion

    elif last_elev <= 0:
        riv_i = riv_i[:-1]
        riv_j = riv_j[:-1]
        course_update = 4  # shortened course

    # if river mouth surrounded by land
    elif low_adj_cell > 0:
        new_riv_i, new_riv_j = find_course(
            z, riv_i, riv_j, len(riv_i), ch_depth, sea_level=sea_level
        )

        new_riv_length = new_riv_i.size - riv_i.size

        if new_riv_length > 0:
            riv_i = new_riv_i
            riv_j = new_riv_j

            if (z[riv_i[-1], riv_j[-1]] - sea_level) < (0.001 * max_cell_h):
                z[riv_i[-1], riv_j[-1]] = (0.001 * max_cell_h) + sea_level

            downcut.cut_new(
                riv_i[-(new_riv_length + 2) :],
                riv_j[-(new_riv_length + 2) :],
                z,
                sea_level,
                ch_depth,
                dx=dx,
                dy=dy,
            )

            course_update = 6  # lengthened land-locked course

        else:
            riv_i = riv_i
            riv_j = riv_j

    # if river mouth needs to prograde
    elif last_elev >= max_cell_h:
        sorted_n = sort_lowest_neighbors(test_elev, (riv_i[-1], riv_j[-1]))
        subaerial_loc = np.where(test_elev[sorted_n] > 0)

        if len(subaerial_loc[0]):
            subaerial_cells = sorted_n[0][subaerial_loc], sorted_n[1][subaerial_loc]

            if (subaerial_cells[0][0], subaerial_cells[1][0]) not in zip(riv_i, riv_j):
                riv_i = np.append(riv_i, subaerial_cells[0][0])
                riv_j = np.append(riv_j, subaerial_cells[1][0])

                if (z[riv_i[-1], riv_j[-1]] - sea_level) < (0.001 * max_cell_h):
                    z[riv_i[-1], riv_j[-1]] = (0.001 * max_cell_h) + sea_level

                # line below not needed if downcutting
                # z[riv_i[-1], riv_j[-1]] -= ch_depth

                downcut.cut_new(
                    riv_i[-3:], riv_j[-3:], z, sea_level, ch_depth, dx=dx, dy=dy
                )

                course_update = 5

            else:
                riv_i = riv_i
                riv_j = riv_j

        else:
            riv_i = riv_i
            riv_j = riv_j

    else:
        riv_i = riv_i
        riv_j = riv_j

    return riv_i, riv_j, course_update



#! /usr/local/bin/python

import numpy as np

from .avulsion_utils import find_beach_length_riv_cell, get_channel_distance


def solve_second_derivative(x, y):
    """Solve the second derivative of y with respect to x.

    Use finite difference to solve the second derivative of *y* with respect
    to *x* where *x* can be unevenly spaced.

    Examples
    --------
    Values can be evenly spaced.

    >>> import numpy as np
    >>> x = np.array([2., 3., 4.])
    >>> y = np.array([4., 9., 16.])
    >>> solve_second_derivative(x, y)
    array([ 2.])

    Values are unevenly spaced.

    >>> x = np.array([2., 3., 5.])
    >>> y = np.array([4., 9., 25.])
    >>> solve_second_derivative(x, y)
    array([ 2.])
    """
    x2_minus_x1 = x[1:-1] - x[:-2]
    x3_minus_x2 = x[2:] - x[1:-1]
    x3_minus_x1 = x[2:] - x[:-2]

    return 2 * (
        y[:-2] / (x2_minus_x1 * x3_minus_x1)
        - y[1:-1] / (x3_minus_x2 * x2_minus_x1)
        + y[2:] / (x3_minus_x2 * x3_minus_x1)
    )


def smooth_rc(dx, dy, nu, dt, ch_depth, riv_i, riv_j, n, sea_level, slope):
    """Smooth river channel elevations using the diffusion equation.

    Parameters
    ----------
    dx : float
        Spacing of grid columns.
    dy : float
        Spacing of grid rows.
    nu : float
        Diffusion coefficient.
    dt : float
        Time step (in seconds).
    riv_i : ndarray
        Row indices for the river path.
    riv_j : ndarray
        Column indices for the river path.
    n : ndarray
        2D array of grid elevations.
    """

    beach_len = find_beach_length_riv_cell(
        n,
        (riv_i[-2], riv_j[-2]),
        (riv_i[-1], riv_j[-1]),
        sea_level,
        ch_depth,
        slope,
        dx=dx,
        dy=dy,
    )

    n_river = n[riv_i, riv_j]
    n_river[-1] = sea_level - ch_depth
    s_river = get_channel_distance((riv_i, riv_j), dx=dx, dy=dy)
    s_river[-1] += beach_len

    dn_rc = (nu * dt) * solve_second_derivative(s_river, n_river)

    n[riv_i[1:-1], riv_j[1:-1]] += dn_rc

    return dn_rc


def calc_crevasse_dep(dx, dy, nu, dt, ch_depth, riv_i, riv_j, n, sea_level, slope, loc):
    """Calculate crevasse splay deposition rate."""

    beach_len = find_beach_length_riv_cell(
        n,
        (riv_i[-2], riv_j[-2]),
        (riv_i[-1], riv_j[-1]),
        sea_level,
        ch_depth,
        slope,
        dx=dx,
        dy=dy,
    )

    n_river = n[riv_i, riv_j]
    n_river[-1] = sea_level - ch_depth
    s_river = get_channel_distance((riv_i, riv_j), dx=dx, dy=dy)
    s_river[-1] += beach_len

    dn_rc = (nu * dt) * solve_second_derivative(s_river, n_river)

    # average deposition in 3 cells above 'breach' to find
    # crevasse splay deposition rate
    if len(dn_rc[:loc]) >= 3:
        splay_dep = np.average(dn_rc[loc - 3 : loc])
    else:
        splay_dep = dn_rc[loc - 1]

    if splay_dep < 0:
        splay_dep = 0

    return splay_dep



#! /usr/local/bin/python
from .avulsion_utils import find_beach_length_riv_cell, get_link_lengths


def calc_qs(nu, riv_i, riv_j, n, sea_level, ch_depth, dx, dy, dt, slope):
    """Calculate sediment flux at river mouth."""

    beach_len = find_beach_length_riv_cell(
        n,
        (riv_i[-2], riv_j[-2]),
        (riv_i[-1], riv_j[-1]),
        sea_level,
        ch_depth,
        slope,
        dx=dx,
        dy=dy,
    )

    ds = get_link_lengths((riv_i[-2:], riv_j[-2:]), dx=dx, dy=dy)
    ds[-1] += beach_len
    dz = (sea_level - ch_depth) - n[riv_i[-2], riv_j[-2]]

    return -nu * dz / ds

    # ds = get_link_lengths((riv_i[-3:-1], riv_j[-3:-1]), dx=dx, dy=dy)

    # dz = n[riv_i[-2], riv_j[-2]] - n[riv_i[-3], riv_j[-3]]

    # return - nu * dz / ds
